"""
Brython-Server main module with Flask route points.
Author: E Dennison
"""
import os
import os.path
import urllib.request
import urllib.parse
import json
import base64
import re
from flask import (
    Flask,
    render_template,
    session,
    request,
    redirect,
    url_for,
    abort,
    Response,
)
import ggame.__version__
from ggame.__version__ import VERSION as GGVERSION, BUZZ_VERSION, PIXI_VERSION
from .reverseproxied import ReverseProxied
from .definitions import (
    ENV_FLASKSECRET,
    ENV_DEBUG,
    ENV_GOOGLECLIENTID,
    ENV_GOOGLEAPIKEY,
    ENV_GOOGLEAPPID,
    BRYTHON_FOLDER,
    BRYTHON_JS,
    ENV_SITETITLE,
    ENV_SITECONTACT,
    ENV_SITEURL,
    INIT_CONTENT,
    RUN_EDIT,
    AUTH_REQUEST,
    AUTH_FORGET,
    IMPORTNAME,
    SESSION_GITHUBCONTEXT,
    SESSION_MAINSHA,
    SESSION_MAINFILE,
    SESSION_METADATA,
    Context,
)
from .utility import (
    githubloggedin,
    checkgithubstate,
    gistrequest,
    githubretrievetoken,
    githubretrievefile,
    githubauthurl,
    githubforgetauth,
    githubgetmainfile,
    githubretrievegist,
    githubpath,
    githubrequest,
)
from .__version__ import VERSION

APP = Flask(__name__, static_url_path="/__static")
APP.wsgi_app = ReverseProxied(APP.wsgi_app)

APP.secret_key = os.environ.get(ENV_FLASKSECRET, "A0Zr98j/3yX R~XHH!jmN]LWX/,?RT")
APP.debug = os.environ.get(ENV_DEBUG, False)

# Retrieve Brython Version
with open(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), BRYTHON_FOLDER, BRYTHON_JS)
) as bjs:
    BRYTHON_VERSION = re.search('__BRYTHON__.__MAGIC__.*=.*"(.+)"', bjs.read()).group(1)

# Locate the ggame directory
GGAME_PATH = os.path.dirname(os.path.abspath(ggame.__version__.__file__))

SITETITLE = os.environ.get(ENV_SITETITLE, "Brython Server")
SITECONTACT = os.environ.get(ENV_SITECONTACT, "noone@nowhere.net")
SITEURL = os.environ.get(ENV_SITEURL, "https://runpython.org")
G_CLIENTID = os.environ.get(ENV_GOOGLECLIENTID, "")
G_APIKEY = os.environ.get(ENV_GOOGLEAPIKEY, "")
G_APPID = os.environ.get(ENV_GOOGLEAPPID, "")


@APP.route("/", methods=["POST", "GET"])
def root():
    """Root server URL.

    This default path for the web site is used for a variety of things,
    via voth POST and GET methods.

    With GET, the URL may include an argument list, which is used to load
    a particularl Github repository/file. This is useful for sharing code
    via e-mail or hyperlink. This URL/method is also used as a return
    URL from Github when the user authorizes the application. In this
    case the argument list includes a STATE and access TOKEN.

    With POST, the user has requested a switch to EDIT mode (from exec.html),
    or to login at Github or forget login at Github.

    Returns one of the following:
    index.html -- render template
    exec.html -- render template
    redirect -- to / or github
    """
    cookieconsent = request.cookies.get("cookie_consent") == "true"

    github_loggedin = githubloggedin()

    returnedhtml = None

    if request.method == "GET":
        if "user" in request.args or "gist" in request.args or "fileid" in request.args:
            # Executing an existing file
            user = request.args.get("user", "")
            repo = request.args.get("repo", "")
            branch = request.args.get("branch", "")
            name = request.args.get("name", request.args.get("gist", ""))
            path = request.args.get("path", "")
            fileid = request.args.get("fileid", "")
            returnedhtml = render_template(
                "exec.html",
                user=user,
                repo=repo,
                branch=branch,
                name=name,
                path=path,
                fileid=fileid,
                title=SITETITLE,
                contact=SITECONTACT,
                brythonversion=BRYTHON_VERSION,
                buzzversion=BUZZ_VERSION,
                pixiversion=PIXI_VERSION,
                bsversion=VERSION,
                ggversion=GGVERSION,
                cookieconsent=cookieconsent,
                g_clientid=G_CLIENTID,
                g_apikey=G_APIKEY,
                g_appid=G_APPID,
            )
        elif "code" in request.args and "state" in request.args:
            # Github authorization response - check if valid
            if checkgithubstate(request.args.get("state")):
                githubretrievetoken(request.args.get("code"))
            returnedhtml = redirect(url_for("root"))
        elif "gui_edit" in request.args:
            # Drive UI integration: Edit
            returnedhtml = render_template(
                "index.html",
                edit=request.args.get("gui_edit", ""),
                new="",
                title=SITETITLE,
                contact=SITECONTACT,
                consolesite=SITETITLE + " Console",
                editcontent="",
                github=github_loggedin,
                brythonversion=BRYTHON_VERSION,
                buzzversion=BUZZ_VERSION,
                pixiversion=PIXI_VERSION,
                bsversion=VERSION,
                ggversion=GGVERSION,
                cookieconsent=cookieconsent,
                g_clientid=G_CLIENTID,
                g_apikey=G_APIKEY,
                g_appid=G_APPID,
            )
        elif "gui_new" in request.args:
            # Drive UI integration: New
            returnedhtml = render_template(
                "index.html",
                edit="",
                new=request.args.get("gui_new", ""),
                title=SITETITLE,
                contact=SITECONTACT,
                consolesite=SITETITLE + " Console",
                editcontent="",
                github=github_loggedin,
                brythonversion=BRYTHON_VERSION,
                buzzversion=BUZZ_VERSION,
                pixiversion=PIXI_VERSION,
                bsversion=VERSION,
                ggversion=GGVERSION,
                cookieconsent=cookieconsent,
                g_clientid=G_CLIENTID,
                g_apikey=G_APIKEY,
                g_appid=G_APPID,
            )
        else:
            # Nothing special going on
            returnedhtml = render_template(
                "index.html",
                github=github_loggedin,
                title=SITETITLE,
                contact=SITECONTACT,
                consolesite=SITETITLE + " Console",
                edit="",
                new="",
                editcontent=INIT_CONTENT,
                brythonversion=BRYTHON_VERSION,
                buzzversion=BUZZ_VERSION,
                pixiversion=PIXI_VERSION,
                bsversion=VERSION,
                ggversion=GGVERSION,
                cookieconsent=cookieconsent,
                g_clientid=G_CLIENTID,
                g_apikey=G_APIKEY,
                g_appid=G_APPID,
            )
    elif request.method == "POST":
        if RUN_EDIT in request.form:
            # user is requesting to open a new page with editor
            returnedhtml = render_template(
                "index.html",
                edit=request.form[RUN_EDIT],
                new="",
                title=SITETITLE,
                contact=SITECONTACT,
                consolesite=SITETITLE + " Console",
                editcontent="",
                github=github_loggedin,
                brythonversion=BRYTHON_VERSION,
                buzzversion=BUZZ_VERSION,
                pixiversion=PIXI_VERSION,
                bsversion=VERSION,
                ggversion=GGVERSION,
                cookieconsent=cookieconsent,
                g_clientid=G_CLIENTID,
                g_apikey=G_APIKEY,
                g_appid=G_APPID,
            )
        elif AUTH_REQUEST in request.form:
            # user is requesting authorization from github
            returnedhtml = redirect(githubauthurl())
        elif AUTH_FORGET in request.form:
            # user is requesting to forget our authorization
            githubforgetauth()
            returnedhtml = redirect(url_for("root"))
    if returnedhtml:
        return returnedhtml
    abort(404)
    return "You should never see this!"


@APP.route("/favicon.ico")
def favicon():
    """Return favicon.ico.

    Since web browsers are inclined to request the favicon.ico from the root
    of the web server, we should be able to provide it. Note that this will
    cause a problem if the Github python app has a resource called favicon.ico.
    """
    return APP.send_static_file("favicon.ico")


@APP.route("/brythonconsole")
def brythonconsole():
    """Return template for python/brython console."""
    cookieconsent = request.cookies.get("cookie_consent") == "true"
    return render_template(
        "console.html",
        title=SITETITLE,
        contact=SITECONTACT,
        consolesite=SITETITLE + " Console",
        brythonversion=BRYTHON_VERSION,
        bsversion=VERSION,
        cookieconsent=cookieconsent,
    )


@APP.route("/" + IMPORTNAME + "/<filename>")
def brythonimport(filename):
    """Return static import file

    Add custom importable modules under the static/IMPORTNAME folder.
    """
    return APP.send_static_file(os.path.join(IMPORTNAME, filename))


@APP.route("/legalnotices/<filename>")
def legalnotices(filename):
    """Return legal notice html"""
    return render_template(
        filename + ".html", title=SITETITLE, contact=SITECONTACT, url=SITEURL
    )


@APP.route("/ggame/<path:filename>")
def ggameimport(filename):
    """Return content from the ggame file tree."""
    try:
        with open(os.path.join(GGAME_PATH, filename), "rb") as thefile:
            content = thefile.read()
        if isinstance(content, bytes):
            return Response(content, mimetype="application/octet-stream")
        return Response(content)
    except (FileNotFoundError) as err:
        print(err)
        abort(404)
    return "You should never see this!"


@APP.route("/ggame.py")
def ggame_py():
    """Return a 404 on any attempt to load ggame.py. This is will
    avoid 'wasting time' before searching for modules in the
    ggame package."""
    # ggame is an available package, no ggame.py will be possible
    abort(404)


@APP.route("/<path:filename>")
def file(filename):
    """Return (possibly cached) file for the current Github repo.
    Will look for png match in the local ggame installation as well!
    """
    filename = urllib.request.pathname2url(filename)
    try:
        cx = Context(*session[SESSION_GITHUBCONTEXT])
        content, _sha = githubretrievefile(cx.user, cx.repo, cx.path + "/" + filename)
    except (FileNotFoundError, KeyError, urllib.error.HTTPError) as err:
        try:
            if not filename.endswith((".png", ".PNG")):
                raise FileNotFoundError
            return ggameimport(filename)
        except (FileNotFoundError) as err:
            print(err)
            abort(404)
    if isinstance(content, bytes):
        return Response(content, mimetype="application/octet-stream")
    return Response(content)


## API routes


@APP.route("/api/v1/commit", methods=["PUT"])
def v1_commit():
    """Commit changes in editor to the current main file on Github.

    JSON arguments:
    user -- Github user name
    repo -- Github user's repo name
    path -- path (fragment) to a specific file
    name -- specific file name
    editcontent -- contents of editor on web page
    commitmsg -- commit message

    JSON return:
    success -- True/False
    """
    content = request.json
    user = content.get("user")
    repo = content.get("repo")
    path = content.get("path", "")
    name = content.get("name", "")
    editcontent = content.get("editcontent", "")
    msg = content.get("commitmsg", "")
    if path and not path.endswith(name):
        path += "/" + name
    else:
        path += name
    try:
        metadata = session.get(SESSION_METADATA, "")  # previously loaded a gist?
        if metadata == "":  # default - ordinary repository
            gitrequest, token = githubrequest(user, repo, path, "PUT")
            parameters = {
                "message": msg,
                "content": base64.b64encode(editcontent.encode("utf-8")).decode(
                    "utf-8"
                ),
                "sha": session[SESSION_MAINSHA],
            }
        else:  # this is a gist file name
            gitrequest, token = gistrequest(name, "PATCH")
            parameters = {"files": {metadata: {"content": editcontent}}}
    # pylint: disable=bare-except
    except:
        print("Session expired.")
        return (
            json.dumps(
                {"success": False, "message": "Session expired - reload to continue"}
            ),
            440,
            {"ContentType": "application/json"},
        )
    data = json.dumps(parameters).encode("utf-8")
    gitrequest.add_header("Content-Type", "application/json; charset=utf-8")
    gitrequest.add_header("Accept", "application/json")
    gitrequest.add_header("Content-Length", len(data))
    try:
        response = urllib.request.urlopen(gitrequest, data)
        jsresponse = json.loads(response.read().decode("utf-8"))
        session[SESSION_MAINSHA] = jsresponse.get("content", {}).get("sha", "")
        return (json.dumps({"success": True}), 200, {"ContentType": "application/json"})
    except urllib.error.HTTPError as err:
        error = err.msg + " " + str(err.code)
        print(
            "Github commit error: " + error + ", token was ",
            token,
            ", path was ",
            user,
            repo,
            path,
        )
        return (
            json.dumps({"success": False, "message": error}),
            404,
            {"ContentType": "application/json"},
        )


@APP.route("/api/v1/load", methods=["PUT"])
def v1_load():
    """Load source code and resources from Github

    JSON arguments:
    user -- Github user name (blank for gist)
    repo -- Github user's repo name (blank for gist)
    branch -- optional branch identifier
    path -- optional path (fragment) to a specific file
    name -- optional specific file name or gist ID

    JSON return:
    success -- True/False
    name -- name of main file to execute
    path -- path to main file
    content -- content of main file
    """
    content = request.json
    user = content.get("user", "")
    repo = content.get("repo", "")
    branch = content.get("branch", "")
    path = content.get("path", "")
    name = content.get("name", "")
    mainfile = name
    mainsha = ""
    try:
        if mainfile == "":
            mainfile = githubgetmainfile(user, repo, path)
        else:
            if user != "" and repo != "":  # user, repo, path and mainfile
                maincontent, mainsha = githubretrievefile(
                    user, repo, path + "/" + mainfile
                )
            elif user == "" or repo == "":  # missing user or repo -> must be gist?
                maincontent, mainsha = githubretrievegist(mainfile)
            else:
                raise FileNotFoundError
            # All files read, save primary name and sha
            session[SESSION_MAINFILE] = mainfile
            session[SESSION_MAINSHA] = mainsha
            session[SESSION_GITHUBCONTEXT] = Context(user, repo, path)
            # All files read, return
            return (
                json.dumps(
                    {
                        "success": True,
                        "name": mainfile,
                        "path": githubpath(user, repo, branch, path, mainfile),
                        "content": maincontent,
                    }
                ),
                200,
                {"ContentType": "application/json"},
            )
    except (urllib.error.HTTPError, FileNotFoundError) as err:
        print("Github error: " + err.msg + ", path was ", user, repo, path)
        if err.msg == "Unauthorized":
            return (
                json.dumps(
                    {
                        "success": False,
                        "message": "Unauthorized: please log in to Github.",
                    }
                ),
                401,
                {"ContentType": "application/json"},
            )
        return (
            json.dumps({"success": False, "message": err.msg}),
            404,
            {"ContentType": "application/json"},
        )
    return (
        json.dumps({"success": False, "message": "You should not see this error."}),
        404,
        {"ContentType": "application/json"},
    )



"""
Brython-Server utility functions for Github and session management
Author: E Dennison
"""

import os
import urllib
import json
import base64
import random
import string
from flask import session, url_for
from .definitions import (
    ENV_GITHUBCLIENTID,
    ENV_GITHUBSECRET,
    SESSION_GITHUBSTATE,
    URL_GITHUBAUTHORIZE,
    URL_GITHUBRETRIEVETOKEN,
    SESSION_METADATA,
    SESSION_ACCESSTOKEN,
)


def github_client_id():
    """Retrieve the Github Client ID."""
    return os.environ.get(ENV_GITHUBCLIENTID, "")


def github_client_secret():
    """Retriev the Github Client Secret."""
    return os.environ.get(ENV_GITHUBSECRET, "")


def newgithubstate():
    """Create new Github STATE for login

    Sets a random string as local session state for Github, if not
    already set.

    Return the state string
    """
    state = session.get(SESSION_GITHUBSTATE, "")
    if SESSION_GITHUBSTATE not in session:
        nchars = 20
        state = "".join(
            random.SystemRandom().choice(string.ascii_uppercase + string.digits)
            for _ in range(nchars)
        )
        session[SESSION_GITHUBSTATE] = state
    return state


def getredirecturl():
    """Construct a redirect URL for returning here from Github

    Return the redirect URL
    """
    url = url_for("root", _external=True)
    if url.endswith(":80"):
        url = url[:-3]
    return url


# Public functions


def checkgithubstate(state):
    """Verify correct Github STATE has been returned

    Return True if matching, False otherwise
    """
    return state == session.get(SESSION_GITHUBSTATE, "")


def githubauthurl():
    """Construct a redirect URL TO Github.

    Return the URL
    """
    url = URL_GITHUBAUTHORIZE + "?"
    url += "client_id=" + github_client_id()
    url += "&redirect_uri=" + getredirecturl()
    url += "&scope=repo gist&state=" + newgithubstate()
    print(url)
    return url


def githubretrievetoken(code):
    """Retrieve a user's access token from Github.

    Returns nothing
    """
    gitrequest = urllib.request.Request(URL_GITHUBRETRIEVETOKEN)
    gitrequest.add_header("Accept", "application/json")
    parameters = {
        "client_id": github_client_id(),
        "client_secret": github_client_secret(),
        "code": code,
        "redirect_uri": getredirecturl(),
    }
    data = urllib.parse.urlencode(parameters)
    data = data.encode("utf-8")
    response = urllib.request.urlopen(gitrequest, data)
    jsresponse = json.loads(response.read().decode("utf-8"))
    if "access_token" in jsresponse:
        session[SESSION_ACCESSTOKEN] = jsresponse.get("access_token")


def githubforgetauth():
    """Forget the current Github authorization."""
    session.pop(SESSION_ACCESSTOKEN, None)


def finishrequestsetup(url, method):
    """Complete preparations for initiating a request to Github

    Arguments:
    url -- the github url for the request.
    method -- http mehthod (e.g. GET, etc.)

    Return tuple:
    gitrequest - the request object from urllib.request.Request
    token - the token being used in the request
    """
    token = session.get(SESSION_ACCESSTOKEN)
    gitrequest = urllib.request.Request(url, method=method)
    gitrequest.add_header("Authorization", "token {0}".format(token))
    gitrequest.add_header("User-Agent", "Brython-Server")
    return gitrequest, token


def finishrequest(gitrequest, retrievalmethod, metamethod=None):
    """Boilerplate for finishing the API request to Github

    Arguments:
    gitrequest -- request object
    token -- session token
    retrievalmethod -- function for extracting file contents from response
    metamethod -- optional function for extracting metadata from response

    Return tuple:
    binreturn - the resource data
    sha - the resource SHA
    """
    jsresponse = sha = None
    try:
        response = urllib.request.urlopen(gitrequest)
        jsresponse = json.loads(response.read().decode("utf-8"))
        sha = jsresponse.get("sha", "")
    except (urllib.error.HTTPError) as err:
        if err.code != 304:  # Not Modified - use earlier jsresponse
            raise
    binreturn = retrievalmethod(jsresponse)
    session[SESSION_METADATA] = metamethod(jsresponse) if metamethod else ""

    try:
        return binreturn.decode("utf-8"), sha
    except UnicodeDecodeError:
        return binreturn, sha


def gistrequest(gistid, method="GET"):
    """Initiate a request to the Github gist API.

    Arguments:
    gistid -- the hex identifier for the file
    method -- http mehthod (e.g. GET, etc.)  (Default is 'GET')

    Return tuple:
    gitrequest - the request object from urllib.request.Request
    token - the token being used in the request
    """
    url = "https://api.github.com/gists/{0}".format(gistid)
    return finishrequestsetup(url, method)


def githubrequest(user, repo, path, method="GET"):
    """Initiate a request to the Github content API.

    Arguments:
    user -- the Github user ID/name
    repo -- the Github user's repository name
    path -- optional path fragment or path to file in the repo
    method -- http mehthod (e.g. GET, etc.)  (Default is 'GET')

    Return tuple:
    gitrequest - the request object from urllib.request.Request
    token - the token being used in the request
    """
    url = "https://api.github.com/repos/{0}/{1}/contents/{2}".format(user, repo, path)
    return finishrequestsetup(url, method)


def githubretrievegist(gistid):
    """Retrieve a gist from Github via API.

    Arguments:
    gistid -- hex string identifying a gist

    Return tuple:
    content -- the content of the specific file
    sha -- the file sha (used for subsequent commits, if any)
    """
    gitrequest, _token = gistrequest(gistid)
    return finishrequest(
        gitrequest,
        lambda x: x["files"][list(x["files"].keys())[0]]["content"].encode(
            "utf-8"
        ),  # content
        lambda x: list(x["files"].keys())[0],
    )  # file name


def githubretrievefile(user, repo, path):
    """Retrieve a specific file from Github via API.

    Arguments:
    user -- the Github user ID/name
    repo -- the Github user's repository name
    path -- specific path to file within the repo

    Return tuple:
    content -- the content of the specific file
    sha -- the file sha (used for subsequent commits, if any)
    """
    retrievalmethod = lambda x: base64.b64decode(x["content"].encode("utf-8"))
    gitrequest, _token = githubrequest(user, repo, path)
    return finishrequest(gitrequest, retrievalmethod)


def githubgetmainfile(user, repo, path):
    """Retrieve the 'main' Python file from a repo/directory. Function
    attempts to make a sensible decision.

    Arguments:
    user -- the Github user ID/name
    repo -- the Github user's repository name
    path -- specific directory path within the repo

    Return: the name of the file
    """
    jsresponse = None
    gitrequest, _token = githubrequest(user, repo, path)
    response = urllib.request.urlopen(gitrequest)
    jsresponse = json.loads(response.read().decode("utf-8"))
    names = [f["name"] for f in filter(lambda x: x["type"] == "file", jsresponse)]
    return selectmainfile(names)


def githubpath(user, repo, branch, path, name):
    """Build a valid URL to file on Github.

    Note: This is sensitive to changes in how Github URLs work.

    Arguments:
    user -- the Github user ID/name
    repo -- the Github user's repository name
    branch -- specific branch name (typ. master or main)
    path -- specific path to file within the repo
    name -- specific file name or gist id

    Returns URL to Github file or gist.
    """
    if user != "" and repo != "":
        retval = "https://github.com/{0}/{1}/blob/{2}/".format(user, repo, branch)
    else:
        retval = "https://gist.github.com/"
    if path:
        retval += path
        if name not in path:
            retval += "/" + name
    else:
        retval += name
    return retval


def githubloggedin():
    """Return whether we are logged in to Github (True/False)."""
    github_token = session.get(SESSION_ACCESSTOKEN)
    return github_token


def selectmainfile(names):
    """Determine 'main' python file from a list of candidates.

    Arguments:
    names -- a list of candidate names
    Return: the best candidate name
    """
    mainfile = ""
    for foundname in names:
        if (
            mainfile == ""
            and len(foundname) > 3
            and foundname[-3:] == ".py"
            or
            # Foud a python file called main.py or __main__.py? Make IT the one!
            foundname in ["main.py", "__main__.py"]
        ):
            mainfile = foundname
    return mainfile



"""
Brython-Server turtle import wrapper
Author: E Dennison
"""
from browser import document
from turtle import *
from turtle import set_defaults, FormattedTuple

set_defaults(turtle_canvas_wrapper=document["graphics-column"])

# Redefine done
_done = done


def done():
    _done()
    Screen().reset()
    Turtle._pen = None


# Redefine FormattedTuple to support abs()
_FormattedTuple = FormattedTuple


class FormattedTuple(_FormattedTuple):
    def __abs__(self):
        return (self[0] ** 2 + self[1] ** 2) ** 0.5



"""
Brython-Server string constant definitions.
Author: E Dennison
"""

from collections import namedtuple

ENV_GITHUBCLIENTID = "githubclientid"
ENV_GITHUBSECRET = "githubsecret"
ENV_GOOGLECLIENTID = "googleclientid"
ENV_GOOGLEAPIKEY = "googleapikey"
ENV_GOOGLEAPPID = "googleappid"
ENV_FLASKSECRET = "flasksecret"
ENV_SITETITLE = "sitetitle"
ENV_SITECONTACT = "sitecontact"
ENV_SITEURL = "siteurl"
ENV_DEBUG = "debug"

SESSION_GITHUBSTATE = "githubstate"
SESSION_ACCESSTOKEN = "accesstoken"
SESSION_MAINFILE = "mainfile"
SESSION_MAINSHA = "mainsha"
SESSION_GITHUBCONTEXT = "githubcontext"
SESSION_GITHUBREPO = "githubrepo"
SESSION_GITHUBPATH = "githubpath"
SESSION_METADATA = "metadata"

Context = namedtuple("Context", ["user", "repo", "path"])

RUN_EDIT = "run_edit"
AUTH_REQUEST = "auth_request"
AUTH_FORGET = "auth_forget"
GITHUB_COMMIT = "github_commit"
IMPORTNAME = "brythonserver"

URL_GITHUBAUTHORIZE = "https://github.com/login/oauth/authorize"
URL_GITHUBRETRIEVETOKEN = "https://github.com/login/oauth/access_token"

INIT_CONTENT = 'print("Hello, world.")'

BRYTHON_FOLDER = "static/brython"
BRYTHON_JS = "brython.js"
